import os
import pandas as pd
import numpy as np
from datetime import datetime
from tqdm import tqdm
from scipy import stats
import csv

class DataSanitiser():
    def __init__(self, mode='batch', dir_in=None, dir_out=None, version=None):
        # dir_file:str, dir_config:str, 
        # info
        print("This is a Minimal Viable Product (MVP) only. Future version will consider robustness.")
        self.author = "Colin Li @ 2021"

        # i/o
        self.df_in = None
        self.df_out = None
        self.date_format = "%Y-%m-%d"
        self.version = (
            version,
            datetime.now().strftime(self.date_format)
        )[version is None]

        # formatting
        self.file_format = 'csv'
        self.char_to_esc = [',', '"', "'"]
        self.cols_date = None
        self.cols_percentage = None
        self.cols_float = None
        self.cols_int = None
        self.precision_min = 9


        self.file_format_in = None
        self.header_in_range = 5
        self.footer_in_range = 5
     
        # workload
        self.dir_in = dir_in
        self.dir_out = dir_out
        self.files_xls = []
        self.files_xlsx = []
        self.files_csv = []

        # mapping
        self.df_mapping = None

        # schema
        self.schema = None

        if mode=='batch':
            self.__calculate_workload()
        elif mode=='single':
            pass
        else:
            raise KeyError("Mode should be either 'batch' or 'single'")

    def __calculate_workload(self):
        """
        recursively search relevant data files within a directory
        """
        for root, dirs, fns in os.walk(self.dir_in):
            for fn in fns:
                ext = fn.split('.')[-1]
                if ext.lower()=='xls':
                    self.files_xls.append(os.path.join(root, fn))
                elif ext.lower()=='xlsx':
                    self.files_xlsx.append(os.path.join(root, fn))
                elif ext.lower()=='csv':
                    self.files_csv.append(os.path.join(root, fn))

        print('<Workload>:\n'
              f'Number of Excel File (*.xls ): {len(self.files_xls)}\n'
              f'Number of Excel File (*.xlsx): {len(self.files_xlsx)}\n'
              f'Number of CSV File   (*.csv ): {len(self.files_csv)}\n')

    
    def generate_mapping(self):
        """
        generate mapping file
        """
        file_meta = []
        files_excel = self.files_xlsx + self.files_xls
        files_csv = self.files_csv 
        files = files_excel + files_csv 

        for fdir in tqdm(files):
            fn = os.path.basename(fdir)
            if fdir in files_excel:
                try:
                    ex = pd.ExcelFile(fdir)
                    for sht in ex.sheet_names:
                        for col in ex.parse(sht, nrows=5).columns:
                            file_meta.append((fn, sht, col))
                except ValueError:
                    for col in pd.read_table(fdir, nrows=5).columns:
                        file_meta.append((fn, '', col))
                except:
                    print(f"File '{fn}' cannot be opened. Please mannully add its content to output file!")
                    continue
            elif fdir in files_csv:
                # loop through top 5 rows for header
                i=0
                while i<5:
                    try:
                        columns = pd.read_csv(fdir, header=i, nrows=5).columns
                        break 
                    except:
                        i+=1
                        continue
                for col in columns:
                    file_meta.append((fn, np.nan, col))

        self.df_mapping = pd.DataFrame(file_meta, columns=['fileName', 'sheetName', 'columnName'])
        self.df_mapping['mapToColumnName'] = ''
        self.df_mapping['systemName'] = ''
        self.df_mapping['originalDateFormat'] = ''
        self.df_mapping['leftNChars'] = ''
        self.df_mapping['skipTopN'] = ''
        self.df_mapping['skipLastN'] = ''
        self.df_mapping['noHeader'] = ''
        self.df_mapping.to_excel(os.path.join(self.dir_out, 'mapping_file_' + str(self.version) + '.xlsx'), index=False)
    
    def create_output_df_schema(self, schema):
        self.schema = schema
        self.df_out = pd.DataFrame(self.schema, index=[])
        print("Table has been created!")
        pass

    def detect_header(self):
        pass

    def detect_footer(self):
        pass


    def parse_excel_worksheet(self, dir_file, sheet_name, sys_name, skip_top_n=0, skip_last_n=0, nrows=None, rename_dict=None):
        bn = os.path.basename(dir_file)
        ext = bn.split('.')[-1]
        sheet_name = False if pd.isnull(sheet_name) else sheet_name
        if ext=='xls':
            try:
                df = pd.read_excel(dir_file, sheet_name=sheet_name, header=skip_top_n, skipfooter=skip_last_n, nrows=nrows)
            except ValueError:
                # print(f"Filename: {dir_file} SkipTop: {skip_top_n}")
                df = pd.read_table(dir_file, header=skip_top_n, skipfooter=skip_last_n, engine='python')

        elif ext=='xlsx':
            ex = pd.ExcelFile(dir_file)
            df = ex.parse(sheet_name, header=skip_top_n, skipfooter=skip_last_n, nrows=nrows)
        
        # rename and only include columns in dict
        if rename_dict:
            df.rename(columns=rename_dict, inplace=True)
            df = df[rename_dict.values()]

        df['file_name'] = bn
        df['sys_name'] = sys_name

        return df

    def parse_csv_file(self, dir_file, sys_name, skip_top_n=0, skip_last_n=0, nrows=None, rename_dict=None):
        """
        obsolete!
        """
        bn = os.path.basename(dir_file)
        ext = bn.split('.')[-1]
        if ext.lower()=='csv':
            # print(dir_file, sys_name, skip_top_n, skip_last_n, nrows, rename_dict)
            df = pd.read_csv(dir_file, header=skip_top_n, skipfooter=skip_last_n, nrows=nrows)

        # if bn=='POLISY USER ACCESS REPORT.CSV':
        #     print("Args: ", skip_top_n, skip_last_n, nrows, rename_dict)
        #     display(df.head(10))
        # rename and only include columns in dict
        if rename_dict:
            df.rename(columns=rename_dict, inplace=True)
            df = df[rename_dict.values()]

        df['file_name'] = bn
        df['sys_name'] = sys_name
        

        return df


    def parse_bad_csv(self, dir_file, sys_name, skip_top_n=0, skip_last_n=0, nrows=None, rename_dict=None, miss_header=False):
        
        bn = os.path.basename(dir_file)
        ext = bn.split('.')[-1]
        
        def bad_row_cleaner(row, num_cols, sep="|"):
            return row[:num_cols-1] + [sep.join(str(i) for i in row[num_cols-1:])]
        
        # determine number of columns
        try:
            with open(dir_file) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                num_rows = 0
                len_rec = []
                for row in csv_reader:
                    len_rec.append(len(row))
                    num_rows+=1
                num_cols = stats.mode(len_rec).mode[0]

            # sanitise bad rows
            rows = []
            with open(dir_file) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                for i, row in enumerate(csv_reader):
                    if (len(row)<num_cols) or (i<skip_top_n) or (i>(num_rows-1-skip_last_n)):
                        continue
                    else:
                        row = bad_row_cleaner(row, num_cols)
                        rows.append(row)

            # convert to dataframe
            df = pd.DataFrame(rows)

            if not(miss_header):
                col_names = df.iloc[0] #grab the first row for the header
                df = df[1:] #take the data less the header row
                df.columns = col_names #set the header row as the df header

        except:
            return None

        # rename and only include columns in dict
        if rename_dict:
            df.rename(columns=rename_dict, inplace=True)
            df = df[rename_dict.values()]

        df['file_name'] = bn
        df['sys_name'] = sys_name

        return df

    def truncate_columns(self, df, cols_trunc, nchar):
        for i, col in enumerate(cols_trunc):
            df[col] = df[col].str[:nchar[i]]
        return df

    def validate_mapping_and_parse_data(self, dir_mapping:str, display_df=False):
        
        # update mapping
        self.df_mapping = pd.read_excel(dir_mapping)

        # get excel workload
        files_excel = self.files_xlsx + self.files_xls
        files_csv = self.files_csv
        files = files_excel + files_csv

        # read mapping
        for fdir in files:
            fn = os.path.basename(fdir)
            sheet_names = self.df_mapping[self.df_mapping['fileName']==fn]['sheetName'].unique()
            for sheet_name in sheet_names:
                
                # get details from mapping and exclude unmapped columns
                try:
                    if pd.notnull(sheet_name):
                        t_df = self.df_mapping[
                            (self.df_mapping['fileName']==fn) & \
                            (self.df_mapping['sheetName']==sheet_name) & \
                            (self.df_mapping['mapToColumnName'].notnull())]
                    else:
                        t_df = self.df_mapping[
                            (self.df_mapping['fileName']==fn) & \
                            (self.df_mapping['mapToColumnName'].notnull())]
                    
                    # get file level parsing requirement
                    t_skip_top = 0 if pd.isnull(t_df['skipTopN'].max()) else int(t_df['skipTopN'].max()) 
                    t_skip_last = 0 if pd.isnull(t_df['skipLastN'].max()) else int(t_df['skipLastN'].max())
                    t_sys = t_df['systemName'].max()
                    t_old_cols = t_df['columnName'].tolist()
                    t_new_cols = t_df['mapToColumnName'].tolist()
                    t_no_header = True if "Y" in t_df['noHeader'].tolist() else False
                    # print(fn, t_no_header)

                    # create helper np array for date column sanitisation
                    # print("format is ", t_df['originalDateFormat'].to_numpy())
                    date_idx = np.argwhere(~pd.isnull(t_df['originalDateFormat'].to_numpy())).flatten()
                    t_date_cols = t_df['mapToColumnName'].to_numpy()[date_idx].flatten()
                    t_date_format = t_df['originalDateFormat'].to_numpy()[date_idx].flatten()
                    # print(t_date_cols, t_date_format)
                    # print(fn, sheet_name, t_skip_top, t_skip_last, t_sys, t_old_cols, t_new_cols, t_date_format)

                    trunc_idx = np.argwhere(~pd.isnull(t_df['leftNChars'].to_numpy())).flatten()
                    t_trunc_cols = t_df['mapToColumnName'].to_numpy()[trunc_idx].flatten()
                    t_trunc_nchars = t_df['leftNChars'].to_numpy()[trunc_idx].flatten().astype(int)
                    # print(t_trunc_cols, t_trunc_nchars)

                except:
                    raise KeyError("Please check if fileName and sheetName are correct in the mapping file!")

                # parse data accordingly
                
                rename_dict = {k:v for (k,v) in zip(t_old_cols, t_new_cols)}

                if fdir in files_excel:
                    df_single = self.parse_excel_worksheet(fdir, sheet_name, t_sys, t_skip_top, t_skip_last, None, rename_dict)
                elif fdir in files_csv:
                    # print(fn, rename_dict)
                    # df_single = self.parse_csv_file(fdir, t_sys, t_skip_top, t_skip_last, None, rename_dict)
                    df_single = self.parse_bad_csv(fdir, t_sys, t_skip_top, t_skip_last, None, rename_dict, t_no_header)

                    if df_single is None:
                        df_single = self.parse_csv_file(fdir, t_sys, t_skip_top, t_skip_last, None, rename_dict)

                if display_df:
                    display(df_single.head(3))
                # df_single.to_excel(fn, index=False)
                
                # truncate columns
                df_single = self.truncate_columns(df_single, t_trunc_cols, t_trunc_nchars)

                # convert date to string
                df_single = self.convert_date_to_string(df_single, t_date_cols, t_date_format)

                # print(sheet_name)
                try:
                    self.df_out = self.df_out.append(df_single, ignore_index=True)
                except:
                    print(f"File {fn}: Please make sure there are no duplicated names in 'mapToColumnName' in mapping document")
                
                print(f"Parsed file: {fn:<40} | Sheet: {sheet_name:<40} | Number of Rows: {df_single.shape[0]:9} | Number of Columns: {df_single.shape[1]:5}")
                # display(df_single.head(3))
    
    def convert_date_to_string(self, df, cols_date, cols_date_format):
        """
        convert date to string as per Hadoop requirements
        """
        df = df.copy()
        for col, date_format in zip(cols_date, cols_date_format):
            # print(col, date_format)
            try:
                df[col] = df[col].dt.strftime(self.date_format)
            except:
                # print("Issue Col Name:", col)
                # display(df.head(5))
                df[col] = pd.to_datetime(df[col])
                df[col] = df[col].dt.strftime(self.date_format)

            # except AttributeError:

            #     df[col] = pd.datetime(df[col], format=date_format)
            #     df[col] = df[col].dt.strftime(self.date_format)
        return df
    
    def remove_sensitive_chars(self):
        for col in tqdm(self.df_out.columns):
            if self.df_out[col].dtype=='O':
                self.df_out[col].fillna('', inplace=True)
                self.df_out[col] = self.df_out[col].str.replace(',', '', regex=False)
                self.df_out[col] = self.df_out[col].str.replace('/', '', regex=False)
                self.df_out[col] = self.df_out[col].str.replace('\\', '', regex=False)
                self.df_out[col] = self.df_out[col].str.replace('"', '', regex=False)
                self.df_out[col] = self.df_out[col].str.replace("'", '', regex=False)

    def save(self):
        self.df_out.to_csv(os.path.join(self.dir_out, "user_access_listing_"+str(self.version)+".csv"), index=False)

    # def parse(self, dir_data:str, dir_config:str):
    #     # load
    #     self.file_format_in = dir_file.split(".")[-1]
    #     if self.file_format_in == 'csv':
    #         try:
    #             self.df_in = pd.read_csv()
    #         except:
    #             print("Error")
    #     elif self.file_format_in in ['xlsx', 'xls']:
    #         try:
    #             self.df_in = pd.read_excel()
    #         except:
    #             print("Error")
    #     else:
    #         print(f"File format '{self.file_format_in}' is not supported!")
    #         return None